# include <Arduino.h>
# include "ADC.h"

volatile char databuf[ ADCDATABUFSIZE ];
volatile unsigned int index = 0;
volatile unsigned long samples = 0;
volatile bool triggered = false;
volatile unsigned char old_val = 0;
volatile unsigned char before_trigger = 0;
volatile unsigned char trigger_val = 0;
volatile int running_average = 0;
volatile int low_pass = 0;
volatile unsigned int running_min = 255;
volatile unsigned int running_max = 0;

ISR( ADC_vect )
{
  samples++;
  if( triggered == false && ADCH > old_val && ADCH > trigger_val && ADCH < trigger_val + 20 )
  {
    triggered = true;
    index = 0;
    before_trigger = old_val;
  }

  int ival = int(ADCH) << 4;
  low_pass = low_pass + ((ival - low_pass) >> 4);
  running_average = low_pass;
  // running_average = (running_average * (ADCDATABUFSIZE-1) + ADCH) / ADCDATABUFSIZE;
  // running_average += (ADCH - running_average) / 32;
  // running_average -= running_average / 32;
  // running_average += ADCH / 32;

  if( ADCH < running_min )
    running_min = ADCH;
  if( ADCH > running_max )
    running_max = ADCH;

  old_val = ADCH;

  if( triggered && index < ADCDATABUFSIZE-1 )
  {
    databuf[ index++ ] = ADCH;
  }

  if( triggered )
    old_val = 255;
}

void
ADC_rate( unsigned prescale )
{
  ADCSRA &= ~(1<<ADEN);                           // stop
  ADCSRA = 0b00111000 + (prescale & 7);           // set rate
  ADCSRA |= (1<<ADEN) | (1<<ADSC);                // start back up
}

unsigned char
ADC_rate( void )
{
  return ADCSRA & 7;
}

void
ADC_init( unsigned char pin )
{
  // The ADMUX Register:
  // • Bit 7:6 – REFS[1:0]: Reference Selection Bits
  //   These bits select the voltage reference for the ADC. If these bits are
  //   changed during a conversion, the change will not go in effect until this
  //   conversion is complete (ADIF in ADCSRA is set). The internal voltage
  //   reference options may not be used if an external reference voltage is
  //   being applied to the AREF pin.
  // • Bit 5 – ADLAR: ADC Left Adjust Result
  //   The ADLAR bit affects the presentation of the ADC conversion result
  //   in the ADC Data Register. Write one to ADLAR to left adjust the result.
  //   Otherwise, the result is right adjusted. Changing the ADLAR bit will
  //   affect the ADC Data Register immediately, regardless of any ongoing
  //   conversions.
  // • Bit 4 – Reserved
  //   This bit is an unused bit in the ATmega48A/PA/88A/PA/168A/PA/328/P,
  //   and will always read as zero.
  // • Bits 3:0 – MUX[3:0]: Analog Channel Selection Bits
  //   The value of these bits selects which analog inputs are connected to
  //   the ADC. If these bits are changed during a conversion, the change
  //   will not go in effect until this conversion is complete
  //   (ADIF in ADCSRA is set)

  // Reference is external AREF pin, pin is A0
  // enum { ADCMultiplexorRegisterInitialValue = 0b00100000 };

  // Reference is 2.56 volts internal reference, pin is A0
  // enum { ADCMultiplexorRegisterInitialValue = 0b11100000 };

  // Reference is VCC volts, pin is A0
  enum { ADCMultiplexorRegisterInitialValue = 0b01100000 };
  
  ADMUX = ADCMultiplexorRegisterInitialValue + (pin & 7);

  // The ADCSRA Register:
  // • Bit 7 – ADEN: ADC Enable
  //   Writing this bit to one enables the ADC. By writing it to zero, the
  //   ADC is turned off. Turning the ADC off while a conversion is in
  //   progress, will terminate this conversion.
  // • Bit 6 – ADSC: ADC Start Conversion
  //   In Single Conversion mode, write this bit to one to start each
  //   conversion. In Free Running mode, write this bit to one to start
  //   the first conversion. The first conversion after ADSC has been
  //   written after the ADC has been enabled, or if ADSC is written at
  //   the same time as the ADC is enabled, will take 25 ADC clock cycles
  //   instead of the normal 13. This first conversion performs
  //   initialization of the ADC. ADSC will read as one as long as a
  //   conversion is in progress. When the conversion is complete, it
  //   returns to zero. Writing zero to this bit has no effect.
  // • Bit 5 – ADATE: ADC Auto Trigger Enable
  //   When this bit is written to one, Auto Triggering of the ADC is
  //   enabled. The ADC will start a conversion on a positive edge of
  //   the selected trigger signal. The trigger source is selected by
  //   setting the ADC Trigger Select bits, ADTS in ADCSRB.
  // • Bit 4 – ADIF: ADC Interrupt Flag
  //   This bit is set when an ADC conversion completes and the Data
  //   Registers are updated. The ADC Conversion Complete Interrupt
  //   is executed if the ADIE bit and the I-bit in SREG are set.
  //   ADIF is cleared by hardware when executing the corresponding
  //   interrupt handling vector. Alternatively, ADIF is cleared by
  //   writing a logical one to the flag. Beware that if doing a
  //   Read-Modify-Write on ADCSRA, a pending interrupt can be disabled.
  //   This also applies if the SBI and CBI instructions are used.
  // • Bit 3 – ADIE: ADC Interrupt Enable
  //   When this bit is written to one and the I-bit in SREG is set,
  //   the ADC Conversion Complete Interrupt is activated.
  // • Bits 2:0 – ADPS[2:0]: ADC Prescaler Select Bits
  //   These bits determine the division factor between the system
  //   clock frequency and the input clock to the ADC.

  // Disable ADC, Start conversion, Auto Trigger Enable, Interrupt Enable,
  // Divide system clock by PRESCALE
  enum { ADControlAndStatusRegisterAInitialValue = 0b00111000 + PRESCALE };
  ADCSRA = ADControlAndStatusRegisterAInitialValue;

  // ADCSRB Register:
  // • Bit 7, 5:3 – Reserved
  //   These bits are reserved for future use. To ensure compatibility
  //   with future devices, these bits must be written to zero when
  //   ADCSRB is written.
  // • Bit 2:0 – ADTS[2:0]: ADC Auto Trigger Source
  //   If ADATE in ADCSRA is written to one, the value of these bits
  //   selects which source will trigger an ADC conversion. If ADATE
  //   is cleared, the ADTS[2:0] settings will have no effect. A
  //   conversion will be triggered by the rising edge of the selected
  //   Interrupt Flag. Note that switching from a trigger source that
  //   is cleared to a trigger source that is set, will generate a
  //   positive edge on the trigger signal. If ADEN in ADCSRA is set,
  //   this will start a conversion. Switching to Free Running mode
  //   (ADTS[2:0]=0) will not cause a trigger event, even if the ADC
  //   Interrupt Flag is set.

  // All bits clear -- free running mode
  enum { ADControlAndStatusRegisterBInitialValue = 0b00000000 };
  ADCSRB = ADControlAndStatusRegisterBInitialValue;

  // DIDR0 Register:
  // • Bits 7:6 – Reserved
  //   These bits are reserved for future use. To ensure compatibility
  //   with future devices, these bits must be written to zero when
  //   DIDR0 is written.
  // • Bit 5:0 – ADC5D...ADC0D: ADC5...0 Digital Input Disable
  //   When this bit is written logic one, the digital input buffer
  //   on the corresponding ADC pin is disabled. The corresponding
  //   PIN Register bit will always read as zero when this bit is set.
  //   When an analog signal is applied to the ADC5...0 pin and the
  //   digital input from this pin is not needed, this bit should be
  //   written logic one to reduce power consumption in the digital
  //   input buffer. Note that ADC pins ADC7 and ADC6 do not have
  //   digital input buffers, and therefore do not require Digital Input
  //   Disable bits.

  // We disable all the digital input buffers
  enum { DigitalInputDisableRegister0InitialValue = 0b00111111 };
  DIDR0 = DigitalInputDisableRegister0InitialValue;

  PRR &= ~(1 << PRADC);       // Disable ADC power reduction
}

void
ADC_start( void )
{
  // Enable ADC and start conversion
  ADCSRA |= (1<<ADEN) | (1<<ADSC);
}

void
ADC_stop( void )
{
  // Disable the ADC
  ADCSRA &= ~(1<<ADEN);
}
