# include <Adafruit_GFX.h>    // Core graphics library
# include <Adafruit_ST7735.h> // Hardware-specific library
# include <SPI.h>
# include "ADC.h"
//# include <Bounce2.h>
# include "RotaryMenu.h"

enum { TFT_CS = 10, TFT_RST = 9, TFT_DC = 8 };

Adafruit_ST7735 tft = Adafruit_ST7735( TFT_CS,  TFT_DC, TFT_RST );

RotaryMenu *rm;

extern struct tp_menu top_menu;

struct trig_menu : public menu
{
  void init( menu *n, menu *c, const char *t )
  {
    next = n;
    child = c;
    text = t;
  }

  void entered( void )
  {
    rm->position( trigger_val );
  }

  void track( bool, long p )
  {
    trigger_val = p;
  }

  menu *clicked( void )
  {
    selected->entered();
    return selected;
  }

  bool allow_free_position( void ) { return true; }
} trigger_menu;

struct spd_menu : public menu
{
  void init( menu *n, menu *c, const char *t )
  {
    next = n;
    child = c;
    text = t;
  }

  void entered( void ) {}

  void track( bool, long ) {}

  menu *clicked( void )
  {
    selected->entered();
    return selected;
  }
} speed_menu;

//
// Anything faster than 1 MHz interrupts so often that little else gets done.
// It does not improve the sample rate, because the processor can't keep up.
//
// struct Hz2000_menu : public menu
// {
//   void init( menu *n, menu *c, const char *t )
//   {
//     next = n;
//     child = c;
//     text = t;
//   }
//
//   void entered( void ) { ADC_rate( PRESCALE_8 ); }
//
//   void track( bool b, long p ) {}
//
//   menu *clicked( void )
//   {
//     selected->entered();
//     return selected;
//   }
// } Hz2000k_menu;

//
// Even at 1 MHz, we don't actually get the full 80,000 samples per second
// we should be getting.
//
struct Hz1000_menu : public menu
{
  void init( menu *n, menu *c, const char *t )
  {
    next = n;
    child = c;
    text = t;
  }

  void entered( void ) { ADC_rate( PRESCALE_16 ); }

  void track( bool, long ) {}

  menu *clicked( void )
  {
    selected->entered();
    return selected;
  }
} Hz1000k_menu;

struct Hz500_menu : public menu
{
  void init( menu *n, menu *c, const char *t )
  {
    next = n;
    child = c;
    text = t;
  }

  void entered( void ) { ADC_rate( PRESCALE_32 ); }

  void track( bool, long ) {}

  menu *clicked( void )
  {
    selected->entered();
    return selected;
  }
} Hz500k_menu;

struct Hz250_menu : public menu
{
  void init( menu *n, menu *c, const char *t )
  {
    next = n;
    child = c;
    text = t;
  }

  void entered( void ) { ADC_rate( PRESCALE_64 ); }

  void track( bool, long ) {}

  menu *clicked( void )
  {
    selected->entered();
    return selected;
  }
} Hz250k_menu;

struct Hz125_menu : public menu
{
  void init( menu *n, menu *c, const char *t )
  {
    next = n;
    child = c;
    text = t;
  }

  void entered( void ) { ADC_rate( PRESCALE_128 ); }

  void track( bool, long ) {}

  menu *clicked( void )
  {
    selected->entered();
    return selected;
  }
} Hz125k_menu;

struct tp_menu : public menu
{
  long previous;
  void init( menu *n, menu *c, const char *t )
  {
    next = n;
    child = c;
    text = t;
    previous = 0;
  }
  void entered( void )
  {
    tft.fillScreen( ST7735_BLACK );
    rm->position( previous );
  }
  void track( bool, long ) {}
  menu *clicked( void )
  {
    previous = rm->position();
    selected->entered();
    return selected;
  }
} top_menu;



ISR( PCINT1_vect )
{
  rm->rotated();
}

ISR( PCINT2_vect )
{
  rm->pushed();
}

RotaryMenu the_rm( 3, A2, A3, &top_menu, tft );

void
setup( void )
{
  top_menu.init( 0, &trigger_menu, "Main" );
  // Hz2000k_menu.init( &Hz1000k_menu, &speed_menu, "160K" );
  Hz1000k_menu.init( &Hz500k_menu, &speed_menu, "80K" );
  Hz500k_menu.init( &Hz250k_menu, &speed_menu, "40K" );
  Hz250k_menu.init( &Hz125k_menu, &speed_menu, "20K" );
  Hz125k_menu.init( &top_menu, &speed_menu, "10K" );
  speed_menu.init( 0, &Hz1000k_menu, "Samples/s" );
  trigger_menu.init( &speed_menu, &top_menu, "Trigger" );

  top_menu.count_children();
  trigger_menu.count_children();
  speed_menu.count_children();
  Hz500k_menu.count_children();
  Hz250k_menu.count_children();
  Hz125k_menu.count_children();

  rm = &the_rm;
  pinMode( 2, OUTPUT );
  digitalWrite( 2, HIGH );

  Serial.begin( 115200 );
  SPI.setClockDivider( SPI_CLOCK_DIV2 );    // 8MHz -- the fastest we can go
  tft.initR( INITR_SIMONTAB );
  tft.fillScreen( ST7735_BLACK );
  tft.setRotation( 1 );
  tft.setTextColor( ST7735_WHITE, ST7735_BLACK );

  sei();                                // Enable interrupts

  ADC_init( 0 );
  ADC_rate( PRESCALE_32 );
  ADC_start();

  delay( 500 );
  trigger_val = (running_max - running_min) / 2;
  Serial.print( "Trigger: " );
  Serial.println( trigger_val );
}

unsigned char buf[ 160 ];
unsigned char last_trigger = 0;
long last_position = 0;
unsigned int old_min = 0;
unsigned int old_max = 0;
int old_avg = 0;
unsigned int old_trg = 0;
// double running_average_frequency = 0;
// unsigned int running_count = 0;

void
loop( void )
{
  rm->show_menu();

  if( running_min != old_min || running_max != old_max
    || running_average != old_avg || trigger_val != old_trg )
  {
    tft.setCursor( 0, 16 );
    tft.print( running_min, 10 );
    tft.print( " to " );
    tft.print( running_max, 10 );
    tft.print( " avg:" );
    tft.print( running_average / 32, 10 );
    tft.print( "." );
    tft.print( running_average % 32, 10 );
    tft.print( ", trg=" );
    tft.print( trigger_val, 10 );
    tft.println( "  " );
    old_min = running_min;
    old_max = running_max;
    old_avg = running_average;
    old_trg = trigger_val;
  }

  // Serial.println( buf[0] );
  
  if( triggered && index >= ADCDATABUFSIZE-1 )
  {
    tft.drawLine( 0, buf[ 0 ], 1, buf[ 1 ], ST7735_BLACK );

    //
    // Draw a green line that shows the trigger point
    //
    if( last_trigger != trigger_val )
      tft.drawLine( 0, 127 - last_trigger, 159, 127 - last_trigger, ST7735_BLACK );
    last_trigger = trigger_val;
    tft.drawLine( 0, 127 - last_trigger, 159, 127 - last_trigger, ST7735_GREEN );

    buf[ 0 ] = 127 - (databuf[ 0 ] & 0x7F);

    for( unsigned int x = 1; x < sizeof buf; x++ )
    {
      if( x < sizeof buf )
        tft.drawLine( x, buf[ x ], x + 1, buf[ x + 1 ], ST7735_BLACK );
      buf[ x ] = 127 - (databuf[ x ] & 0x7F);

      tft.drawLine( x - 1, buf[ x - 1 ], x, buf[ x ], ST7735_WHITE );
    }
    triggered = false;
    old_val = 255;
    running_min = 255;
    running_max = 0;
    delay( 1 );
  }
}
