# ifndef Rotary_Menu_h
# define Rotary_Menu_h

# include <Arduino.h>
# include <Adafruit_GFX.h>    // Core graphics library
# include <Adafruit_ST7735.h> // Hardware-specific library

enum { DETENT = 3, LOCKOUT = 4 };

const char
direction_table[] =
{
   0, -1,   1,   0,
   1,  0,   0,  -1,
  -1,  0,   0,   1,
   0,  1,  -1,   0
};

class menu
{
protected:
  menu *next;
  menu *child;
  menu *selected;
  const char *text;
  unsigned char num_children;
  bool erase_me;
public:
  virtual void init( menu *n, menu *c, const char *t ) = 0;
  virtual void entered( void );
  virtual void track( bool, long ) = 0;
  virtual menu *clicked( void ) = 0;
  const char *get_text( void ) { return text; }
  menu *get_child( void ) { return child; }
  menu *get_next( void ) { return next; }
  void set_selected( menu *s ) { selected = s; }
  menu *get_selected( void ) { return selected; }
  unsigned char get_num_children( void ) { return num_children; }
  bool erase( void ) { return erase_me; }
  void erase( bool b ) { erase_me = b; }
  virtual bool allow_free_position( void ) { return false; }
  void count_children( void )
  {
    num_children = 0;
    for( menu *c = child; c; c = c->next )
      num_children++;
    erase_me = true;
  }
};

class RotaryMenu
{
  bool button_state, old_state, changed;
  unsigned char buttonbit, bit1, bit2;
  volatile unsigned char *buttonreg, *reg1, *reg2, last;
  unsigned long old_millis;
  menu *current_menu;
  Adafruit_ST7735 &tft;

  //
  // As the rotary encoder turns, one click causes several changes,
  // adding +1 and -1 sometimes several times. In the end, one click
  // increments by two, so we divide by 2 every time we see the
  // DETENT state.
  //
  long twice_pos, actual_pos;

public:
  RotaryMenu( char but, char p1, char p2, menu *m, Adafruit_ST7735 &t ) : tft( t )
  {
    current_menu = m;
    buttonbit = digitalPinToBitMask( but );
    buttonreg = portInputRegister( digitalPinToPort( but ) );
    bit1 = digitalPinToBitMask( p1 );
    reg1 = portInputRegister( digitalPinToPort( p1 ) );
    bit2 = digitalPinToBitMask( p2 );
    reg2 = portInputRegister( digitalPinToPort( p2 ) );

    pinMode( but, INPUT_PULLUP );
    pinMode( p1, INPUT_PULLUP );
    pinMode( p2, INPUT_PULLUP );

    last = 3;
    twice_pos = 0;
    actual_pos = 0;
    old_millis = millis();
    button_state = false;

    // Enable the pin interrupt, clear any interrupts, enable interrupts
    *digitalPinToPCMSK( but ) |= bit( digitalPinToPCMSKbit( but ) );
    PCIFR  |= bit (digitalPinToPCICRbit( but ));
    PCICR  |= bit (digitalPinToPCICRbit( but ));

    *digitalPinToPCMSK( p1 ) |= bit( digitalPinToPCMSKbit( p1 ) );
    PCIFR  |= bit (digitalPinToPCICRbit( p1 ));
    PCICR  |= bit (digitalPinToPCICRbit( p1 ));

    *digitalPinToPCMSK( p2 ) |= bit( digitalPinToPCMSKbit( p2 ) );
    PCIFR  |= bit (digitalPinToPCICRbit( p2 ));
    PCICR  |= bit (digitalPinToPCICRbit( p2 ));
    changed = true;
  }

  long position()
  {
    return actual_pos;
  }

  void position( long p )
  {
    actual_pos = p;
    twice_pos = p * 2;
  }

  bool
  button( void )
  {
    return button_state;
  }

  menu *
  get_menu( void )
  {
    return current_menu;
  }

  void pushed( void )
  {
    Serial.println( "Pushed" );
    changed = true;
    bool test = (*buttonreg & buttonbit) ? false : true;
    //
    // Read the bit, and then ignore any further transitions for LOCKOUT
    // milliseconds. Thus we ignore any bounces the switch makes.
    //
    if( millis() - old_millis > LOCKOUT )
    {
      button_state = test;
      old_millis = millis();
      if( button_state )
      {
        current_menu = current_menu->clicked();
        current_menu->erase( true );
      }
    }
  }

  void rotated( void )
  {
    changed = true;
    int change1 = (*reg1 & bit1) ? 1 : 0;
    int change2 = (*reg2 & bit2) ? 2 : 0;
    char current = change2 + change1;

    if( current != last )
    {
      twice_pos += direction_table[ last * 4 + current ];
      last = current;

      if( last == DETENT )
        actual_pos = twice_pos / 2;

      Serial.print( "Track: button_state=" );
      Serial.print( button_state );
      Serial.print( ", actual_pos=" );
      Serial.println( actual_pos );

      current_menu->track( button_state, actual_pos );
    }
  }

  void
  show_menu( void )
  {
    if( changed == false )
      return;

    menu *m = get_menu();

    if( m->erase() )
    {
      tft.fillScreen( ST7735_BLACK );
      m->erase( false );
    }

    tft.setCursor( 0, 0 );
    tft.setTextColor( ST7735_GREEN, ST7735_BLACK );
    tft.println( m->get_text() );

    if( m->allow_free_position() == false )
    {
      if( position() >= m->get_num_children() )
        position( m->get_num_children() - 1 );
      if( position() < 0 )
        position( 0 );
    }

    int x = 0;
    m->set_selected( m->get_child() );
    for( menu *c = m->get_child(); c; c = c->get_next() )
    {
      if( position() == x++ || m->allow_free_position() )
      {
        m->set_selected( c );
        tft.setTextColor( ST7735_BLACK, ST7735_WHITE );
      }
      else
        tft.setTextColor( ST7735_WHITE, ST7735_BLACK );
      tft.print( c->get_text() );
      tft.setTextColor( ST7735_WHITE, ST7735_BLACK );
      tft.print( " " );
    }
    tft.println();

    changed = false;
  }

};

# endif // Rotary_Menu_h
